#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sysexits.h>

#include <strfunc.h>

#include <desire.h>

extern char *optarg;
extern int optind;

void usage(char *);

char *service=NULL;
char *name=NULL;
char *config=NULL;

int main(int ac, char **av) {
	int ch;
	ctb_range range;
	ctb_data *data;
	ctb_session *ctbs;
	int count;

	memset(&range, 0, sizeof(ctb_range));
	range.either=1;
	range.simple_def = 0;

	while((ch=getopt(ac, av, "s:n:f:hmdwyclp1:2:")) != -1) {
		switch(ch) {
		case 's': service=optarg; break;
		case 'n': name=optarg; break;
		case 'f': config=optarg; break;

		case 'h': range.simple_def |= HOUR; break;
		case 'd': range.simple_def |= DAY; break;
		case 'w': range.simple_def |= WEEK; break;
		case 'm': range.simple_def |= MONTH; break;
		case 'y': range.simple_def |= YEAR; break;

		case 'c': range.simple_def |= CURR; break;
		case 'l': range.simple_def |= LAST; break;
		case 'p': range.simple_def |= PREV; break;

		case '1':	range.either = 0;
					range.from=strtotime(optarg);
					if(!range.to) range.to = time(NULL);
					break;
		case '2':	range.either = 0;
					range.to=strtotime(optarg);
					if(!range.from) range.from = range.to - 86400;
					break;

		default:
			usage(*av);
		};
	};

	if(!range.simple_def)
		range.simple_def = LAST | WEEK;

	if(!service)
		usage(*av);

	if(!(ctbs=ctb_open(config, service, name))) {
		fprintf(stderr, "Can't open CTB session.\n");
		exit(2);
	};

	ac-=optind;
	av+=optind;

	data=ctb_get(ctbs, &range, &count);
	if(!data || count < 0) {
		fprintf(stderr, "Error getting data.\n");
		exit(1);
	};

	fprintf(stderr, "From: %s", ctime((time_t *)&range.from));
	fprintf(stderr, "Till: %s", ctime((time_t *)&range.to));
	fprintf(stderr, "Count: %d\n", count);

	if(!count)
		fprintf(stderr, "No data for this period available.\n");

	while(count--) {
		printf("%s",
			ctb_format(ctbs, "Date:\t%T\n\tFlags:\t%F (%f) ", data)
		);
		printf("[%x]\n %s %u/%u\n",
			data->FL.i,
			ctb_format(ctbs, "\tIn/Out:\t%i/%o\n\tSrc/Dt:\t%s/%d\t(%S/%D)\n\tU1/U2:\t%1/%2", data),
			data->U1, data->U2);

		if(count)
			printf("%d'th:\n", count);
		data++;
	};


	return 0;
};

void
usage(char *pn) {
	fprintf(stderr, "Usage: %s -s <service> [-n <name>] [-f <config>] [{-h|-d|-w|-m|-y}]\n\t[{-c|-l|-p}] [-1 <time_start>] [-2 <time_end>]\n", pn);
	exit(EX_USAGE);
};


