#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sysexits.h>

#include <desire.h>

extern char *optarg;
extern int optind;

void usage(char *);

char *service=NULL;
char *config=NULL;

int printout(unsigned int, char const *, void *);

int main(int ac, char **av) {
	int ch;
	ctb_session *ctbs;
	int field = -1;
	char *pattern="*";

	while((ch=getopt(ac, av, "s:f:FDSIO12")) != -1) {
		switch(ch) {
		case 's': service=optarg; break;
		case 'f': config=optarg; break;
		case 'F':
			field = FIELD_FL; break;
		case 'S':
			field = FIELD_SRC; break;
		case 'D':
			field = FIELD_DST; break;
		case 'I':
			field = FIELD_IN; break;
		case 'O':
			field = FIELD_OUT; break;
		case '1':
			field = FIELD_U1; break;
		case '2':
			field = FIELD_U2; break;
		default:
			usage(*av);
		};
	};

	if(!service || field < 0)
		usage(*av);

	ac -= optind;
	av += optind;

	if(!(ctbs=ctb_open(config, service, NULL))) {
		fprintf(stderr, "Can't open CTB session.\n");
		exit(2);
	};

	if(ac)
		pattern = *av;

	fprintf(stderr, "Fuzzy pattern: %s\n", pattern);

	ctb_revmap(ctbs, field, pattern, RVM_FUZZY, printout, NULL);


	ctb_close(ctbs);

	return 0;
};

void
usage(char *pn) {
	fprintf(stderr, "Usage: %s [-f <config>] -s <service> {-F|-D|-S|-I|-O|-1|-2} [<pattern>]\n", pn);
	exit(EX_USAGE);
};


int
printout(unsigned int Idx, char const *Value, void *optKey) {
	(void)optKey;

	printf("[%08x] -> [%s]\n", Idx, Value);

	return 0;
};
