/* $Id: pg_expr.c,v 1.2 2001/03/28 10:36:01 vlm Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <string.h>
#include <pwd.h>
#include <ctype.h>
#include <setjmp.h>
#include <signal.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#include <strfunc.h>

#include "desire.h"
#include "pg_expr.h"

expr_t *expressions=NULL;

unsigned char
get_urlre(char *url) {
	register expr_t *expr;
	register char *s;
	register unsigned int value = 0;

	for(expr = expressions; expr; expr=expr->next) {
		if(expr->type == 2) {
			if(!(s=sed_exec(expr->e_sed, url)))
				continue;

			if(expr->result == -1) {
				if(*s < '0' || *s > '9')
				fprintf(stderr,
				"Warning: regular expressions returned non-numeric value.\n");
				expr->result = strtol(s, NULL, 0);

				if(expr->result > 255 || expr->result < 0) {
					fprintf(stderr,
					"Warning: regular expressions returned too big value.\n");
					return 0;
				};
			};

		} else {
			if(!strstr(url, expr->e_char))
				continue;
		};

		value |= (unsigned int)expr->result;

		if(expr->flags == 1)
			return value;
	};

	return value;
};

void
add_expr(expr_t *src) {
	expr_t *dst;

	if((dst=expressions)) {
		while(dst->next) dst=dst->next;
		dst = dst->next = (expr_t *)malloc(sizeof(expr_t));
	} else {
		dst = expressions = (expr_t *)malloc(sizeof(expr_t));
	};

	if(!dst) {
		fprintf(stderr, "Memory allocation failed.\n");
		exit(EX_TEMPFAIL);
	};

	memcpy(dst, src, sizeof(expr_t));
	dst->next = NULL;
};

void
read_expressions(char *fname) {
	FILE *fp;
	slist *sl;
	sbuf *sb;
	sed_t *se;
	expr_t expr;
	char *s;

	if(!(fp=fopen(fname, "r"))) return;

	sb = sbuf_init();
	sl = sinit();

	while((sbuf_zero(sb) == 0) && sbuf_fgets(sb, fp)) {
		if(sb->len < 5 || sb->buf[0] == '#')
			continue;

		sb->buf[--sb->len] = '\0';

		sclear(sl);
		if(splitf(sl, sb->buf, " \t\n", 0) > 1) {

			if(sl->count > 3) {
				fprintf(stderr, "Unknown expression: [%s]\n", sb->buf);
				exit(EX_DATAERR);
				continue;
			};

			expr.e_char = sf_strdup(sl->list[0]);
			for(s = expr.e_char; *s; s++)
				*s = tolower(*s);

			if(sl->count == 3 && eq(sl->list[2], "stop")) {
				expr.flags = 1;
			} else {
				expr.flags = 0;
			};

			expr.type = 0;
			expr.result = atoi(sl->list[1]);

			if(expr.result < 0 || expr.result > 255) {
				fprintf(stderr, "Numeric value out of range: %s\n",
					sl->list[1]);
				exit(EX_DATAERR);
			};

		} else {

			if(sb->buf[sb->len-1] == 's') {
				expr.flags = 1;
				sb->buf[sb->len-1] = '\0';
			} else {
				expr.flags = 0;
			};
	
			se=sed_compile(sb->buf);
			if(!se) {
				fprintf(stderr, "Can't compile regular expression %s\n", sb->buf);
				exit(EX_DATAERR);
			};

			expr.type = 2;
			expr.e_sed = se;
			expr.result = -1;
		};

			add_expr(&expr);
	};


};
