#include <string.h>
#include <strfunc.h>
#include <ctype.h>

#include "pg_url.h"

#define	INVALID_DOMAIN	"INVALID"

int
check_domain_ok(char *url, int len) {
	char *s = url;
	register int hasalpha = 0;
	register int hasdot = 0;

	if(!*s) return NULL;
	if(!(*s >= 'a' && *s <= 'z')
		&& !(*s >= '0' && *s <= '9')) return NULL;

	for(; *s && len; s++, len--) {
		if(
			(*s >= 'a' && *s <= 'z' && (hasalpha=1))
			|| (*s == '.' && (hasdot=1))
			|| (*s >= '0' && *s <= '9')
			|| (*s == '-' && (hasalpha=1) )
			|| (*s == '_' && (hasalpha=1) )
		) continue;
		break;
	};

	if(len > 0)
		return 0;	/* Invalid character */

	if(len < 0 && *s)
		return 0;	/* Invalid character */

	/* Move to the last character */
	s--;

	if(hasalpha && *s == '.')	/* www.dot.com. */
		*s-- = '\0';

	if(hasalpha) {
		if(*s < 'a' || *s > 'z') {
			if(!hasdot) {
				if(*s < '0' || *s > '9')
					return 0;
				return 1;
			};
			return 0;	/* Invalid character at the end of the string */
		};
		return 1;
	};

	if(!hasalpha) {
		if(*s < '0' || *s > '9')
			return 0;	/* Invalid character at the end of the string */
		return 1;
	};

	return 0;
};

struct url *
split_url(char *url) {
	/*
	** up:
	** [PROTO]://[USER[:PASSWD]@][DOMAIN[:PORT]]/[PATH]?[ARGS]
	*/
	static struct url parsed;
	static sbuf *sb_full = NULL;
	static sbuf *sb_proto = NULL;
	static sbuf *sb_user = NULL;
	static sbuf *sb_password = NULL;
	static sbuf *sb_domain = NULL;
	static sbuf *sb_port = NULL;
	static sbuf *sb_path = NULL;
	static sbuf *sb_args = NULL;
	register char *p, *c;

	if(!url || !*url)
		return NULL;

	memset(&parsed, 0, sizeof(parsed));

	if(!sb_full) {
		sb_full = sbuf_init();
		sb_proto = sbuf_init();
		sb_user = sbuf_init();
		sb_password = sbuf_init();
		sb_domain = sbuf_init();
		sb_port = sbuf_init();
		sb_path = sbuf_init();
		sb_args = sbuf_init();
	};

	/*********************/
	/* URL canonization: */
	/*********************/

	/* Url-decoding... */
	if(strchr(url, '%'))
		url = url_decode(url);

	/* Lower-case'ing... */
	for(p = url; *p; p++)
		if( *p >= 'A' && *p <= 'Z')
			*p += 32;

	/****************************/
	/* End of URL canonization: */
	/****************************/



	/* 0. Canonized url itself. */

	sbuf_zero(sb_full);
	sbuf_add(sb_full, url);
	parsed.url = sb_full->buf;

	/* 1. PROTO */

	for( p = url; *p; p++ ) {
		if(*p == ':') {
			*p = 0;

			if( p[1] == '/' && p[2] == '/' ) {

				sbuf_zero(sb_proto);
				sbuf_add(sb_proto, url);
				parsed.proto = sb_proto->buf;
	
				p+=3;
				url = p;
				continue;
			} else if(p[1] >= '0' && p[1] <= '9') {
				for(c = p+1; *c >= '0' && *c <= '9'; c++);

				if(!*c || *c == '/') {	/* localhost:80[/] */

					if(check_domain_ok(url, -1)) {
						sbuf_zero(sb_domain);
						sbuf_add(sb_domain, url);
						parsed.domain = sb_domain->buf;
					} else {
						fprintf(stderr, "Invalid domain: [%s]\n", url);
						parsed.domain = INVALID_DOMAIN;
					};

					sbuf_zero(sb_port);
					sbuf_add2(sb_port, p + 1, c - p - 1);
					parsed.port = sb_port->buf;

					if(!*c)
						return &parsed;

					url = c;
					goto goto_path;
				};

				p = c;
			};

			sbuf_zero(sb_user);
			sbuf_add(sb_user, url);
			parsed.user = sb_user->buf;

			/* skip user:password */
			for(url = p+1; ; url++) {

				if(!*url)
					return NULL;	/* localhost:80sux */

				if(*url == '@') {
					*url = 0;
					sbuf_zero(sb_password);
					sbuf_add(sb_password, p + 1);
					parsed.password = sb_password->buf;
					p = url;
					url++;
					break;
				};
			};
		};
		if(*p == '/') {
			if(url == p)
				return NULL;

			if(check_domain_ok(url, p - url)) {
				sbuf_zero(sb_domain);
				sbuf_add2(sb_domain, url, p - url);
				parsed.domain = sb_domain->buf;
			} else {
				fprintf(stderr, "Invalid domain: [%s]:%d\n", url, p-url);
				parsed.domain = INVALID_DOMAIN;
			};
			
			url = p;
			goto goto_path;
		};
		if(*p == '@') {
			*p = 0;
			sbuf_zero(sb_user);
			sbuf_add(sb_user, url);
			parsed.user = sb_user->buf;
			url = p + 1;
			break;
		}
	};

	/* 2. DOMAIN */

	for( p = url; ; p++ ) {
		if(!*p || *p == '/') {
			*p = '\0';

			if(check_domain_ok(url, -1)) {
				sbuf_zero(sb_domain);
				sbuf_add(sb_domain, url);
				parsed.domain = sb_domain->buf;
			} else {
				fprintf(stderr, "Invalid domain: [%s]\n", url);
				parsed.domain = INVALID_DOMAIN;
			};

			if(!*p)
				return &parsed;

			url = ++p;
			break;
		};
		
	};

	/* 3. PATH */

goto_path:

	for( p = url; ; p++ ) {
		if(!*p || *p == '?') {

			sbuf_zero(sb_path);
			sbuf_add2(sb_path, url, p-url);
			parsed.path=sb_path->buf;

			if(!*p)
				return &parsed;

			url = ++p;
			break;
		};
	};

	/* 4. ARGS */
	sbuf_zero(sb_args);
	sbuf_add(sb_args, url);
	parsed.args=sb_args->buf;

	return &parsed;
};


