/* $Id: proxy_getstat.c,v 1.6 2001/04/09 08:31:51 vlm Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <string.h>
#include <pwd.h>
#include <ctype.h>
#include <setjmp.h>
#include <signal.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#include <strfunc.h>

#ifdef	HAVE_LIBLGBA
#include <lgba.h>

char *usenode=NULL;
#endif

#include "desire.h"

#include "pg_expr.h"
#include "pg_usage.h"
#include "pg_pstat.h"
#include "pg_url.h"

int		get_status(char *);
char	get_way(char *);
char	get_mime(char *mtype, char *path);
char	get_request_method(char *req, char *url);
unsigned int get_who(char *);

char *crop_domain(char *, int level);
unsigned int get_cached_domain_index(char *dmn);
unsigned int compute_domain_index(char *dmn, int Add2Map);
void cache_domain_index(char *dmn, unsigned int);

int recognize_format(FILE *);
size_t process_file(FILE *, int);
int process_request(slist *);

ctb_session *ctbs=NULL;

int do_display_stats = 0;
int do_interrupt = 0;

jmp_buf env;

size_t	file_size=0;
off_t	file_start=0;
off_t	file_pos=0;
size_t	linecount=0;
size_t	lineserr=0;
size_t	linesok=0;
size_t	bytesctb=0;

int ttyon = 1;

void
termdown(int i) {
	(void)i;
	signal(SIGHUP, termdown);
	signal(SIGTTIN, termdown);
	signal(SIGTTOU, termdown);
	if(ttyon == 0) {
		do_display_stats = 1;
		return;
	};
	ttyon = 0;
	freopen("/dev/console", "a", stdout);
	freopen("/dev/console", "a", stderr);
};

void
intstat(int i) {
	(void)i;
	do_display_stats = 1;
	signal(SIGINT, intstat);
};

void
interrupt(int i) {
	(void)i;
	do_interrupt = 1;
	signal(SIGQUIT, SIG_IGN);
};

void
sigabrt(int i) {
	(void)i;
	longjmp(env, 2);
};

int
main(int ac, char **av) {
	int c;
	FILE *fp;
	int pfd;
	char *acco="proxy";
	char *exprfile = NULL;
	char *desconf=NULL;

	signal(SIGABRT, sigabrt);
	signal(SIGHUP, termdown);
	signal(SIGTTIN, termdown);
	signal(SIGTTOU, termdown);
	signal(SIGINT, intstat);
	signal(SIGQUIT, interrupt);

	while(( c = getopt(ac, av, "a:n:e:c:") ) != -1)
	switch(c) {
		case 'a':
			acco = optarg;
			break;
		case 'n':
#ifdef	HAVE_LIBLGBA
			usenode = optarg;
#else
			fprintf(stderr, "Program built without lgba support.\n");
			usage(*av);
#endif
			break;
		case 'c':
			desconf = optarg;
			break;
		case 'e':
			exprfile = optarg;
			break;
		default:
			usage(*av);
	};

	ac -= optind;
	if(ac != 1)
		usage(*av);
	av += optind;

	fp = fopen(*av, "r");
	if(!fp) {
		fprintf(stderr, "Can't open %s.\n", *av);
		exit(EX_NOINPUT);
	};

	if(recognize_format(fp)) {
		fprintf(stderr, "File format of '%s' unrecognized.\n", *av);
		fclose(fp);
		exit(EX_DATAERR);
	};

	if(exprfile) {
		read_expressions(exprfile);
	};

	/* Get the log file size */
	fseek(fp, 0L, SEEK_END);
	file_size = ftell(fp);
	fseek(fp, 0L, SEEK_SET);

	if(!(ctbs=ctb_open(desconf, acco, NULL))) {
		fprintf(stderr, "Can't open DESIRE session.\n");
		exit(EX_UNAVAILABLE);
	};

#ifdef	HAVE_LIBLGBA

	if(usenode) {
		char *avl[] = { "uidnumber", "uid", NULL };
		if(lgba_init(NULL) || lgba_open() || lgba_chnode(usenode)) {
			fprintf(stderr, "Can't initialize LGBA.\n");
			exit(EX_TEMPFAIL);
		};
		lgba_setattrlist(avl);
	};

#endif

	pfd = initFILEpos(*av, fp, "getstat.pos");
	if(pfd == -1) {
		fprintf(stderr, "Can't initialize file position cache...\n");
		exit(EX_UNAVAILABLE);
	};

	/* Get the log file position */
	file_pos = file_start = ftell(fp);

	if( (c=setjmp(env)) == 0 ) {
		process_file(fp, pfd);
	} else {
		/* Abort? Clearing memory!!! */
		if(c == 2) {
			expr_t *expr = expressions;
			expr_t *nextexpr= NULL;
			while(expr) {
				nextexpr = expr->next;
				if(expr->type == 2) {
					sed_free(expr->e_sed);
				} else {
					free(expr->e_char);
				};
				free(expr);
				expr = nextexpr;
			};
		};
	};
	file_pos = ftell(fp);

	for(c = 1; c < 32; c++)
		signal(SIGABRT, SIG_DFL);

	adjFILEpos(fp, pfd, 1);

	ctb_close(ctbs);
	close(pfd);
	fclose(fp);

	fprintf(stderr, "Process completed.\n");
	display_stats();


	return 0;
};


int
recognize_format(FILE *fp) {
	struct stat st;
	sbuf *sb;
	slist *sl;
	char *p;
	int rc = 0;

	if(!fp)
		return -1;

	if(fstat(fileno(fp), &st) == -1)
		return -1;

	if( (st.st_mode & S_IFMT) != S_IFREG )
		return -1;

	sb = sbuf_init();
	p = sbuf_fgets(sb, fp);

	if(!p) {
		sbuf_free(sb);
		return -1;
	};

	sl = split(p, NULL, 0);

	sbuf_free(sb);

	if(sl->count != 10)
		rc = -1;

	sfree(sl);

	rewind(fp);

	return rc;
};

size_t
process_file(FILE *fp, int pfd) {
	sbuf *sb;
	slist *req;
	char *p;
	sigset_t set;
	sigset_t oset;
	int ret;

	sigfillset(&set);
	sigprocmask(SIG_BLOCK, &set, &oset);

	sb = sbuf_init();
	req = sinit();

	while((sbuf_zero(sb) == 0) && (p = sbuf_fgets(sb, fp))) {
		linecount++;

		sclear(req);
		splitf(req, p, " \n\r", 0);
		if(req->count == 9) sadd(req, "-");

		if(req->count != 10) {
			fprintf(stderr, "[Invalid entry format. Split by %d tokens] => [%s]\n", req->count, sjoin(req, "], ["));
			continue;
		};

		ret = process_request(req);
		if(ret == 0)
			linesok++;
		else if(ret == -1) {
			lineserr++;
			printf("\t%s\n", sb->buf);
		};

		/* Remember position every N log lines */
		if((linecount % 1000) == 0) {
			adjFILEpos(fp, pfd, 0);
		};

		if(do_display_stats) {
			do_display_stats = 0;
			file_pos = ftell(fp);
			fprintf(stderr, "\nSoft interrupt received.\nUse SIGQUIT (Ctrl+\\) to hard break.\n");
			display_stats();
		};

		if(do_interrupt)
			break;

		/* Allow signals to come every N lines. */
		if((linecount % 5000) == 0) {
			sigprocmask(SIG_SETMASK, &oset, &set);
			sigfillset(&set);
			sigdelset(&set, SIGINT);
			sigdelset(&set, SIGQUIT);
			sigprocmask(SIG_BLOCK, &set, &oset);
		};

	};


	sfree(req);
	sbuf_free(sb);

	return linecount;
};

ctb_data cd;

/*
TS  Time stamp
FL  Flags	 - [<way>] [<mime>] [<url_contain>] [<flags>]
	way:
		0: NONE/NULL (cache hit)
		1: DIRECT
		2: PROXY
	mime:
		0x0?: text/
			1:	/plain
			2:	/html
		0x1?: image/
			1:	/gif
			2:	/jpg
		0x2?: binary
			0:	/octet-stream & alike
			1:	archives
			2:	documents
			3:	audio/
			4:	video/
	url_contain:
		url substring or regexp
	flags:
		(&0x03)==0x00:	unknown method
		(&0x03)==0x01:	http
		(&0x03)==0x02:	ftp
		(&0x03)==0x03:	connect

In  In		 - File size
Out Out		 - Time spent to serve request
Src Source	 - Requestor IP
Dst Destination	 - tld:8/dmn:24 (TLD hash/full domain)
U1  User defined - User requested (uidnumber)
U2  User defined - ?
*/

#define	tstamp	(req->list[0])	/* .TS */
#define	tspent	(req->list[1])	/* .Out */
#define	tfromip	(req->list[2])	/* .Src */
#define	tstatus	(req->list[3])
#define	tsize	(req->list[4])	/* .In */
#define	treq	(req->list[5])	/* .Fl.c.s2c2 */
#define	turl	(req->list[6])	/* .Fl.c.s2c1 */
#define	twho	(req->list[7])	/* .U1 */
#define	tway	(req->list[8])	/* .Fl.c.s1c1 */
#define	tmime	(req->list[9])	/* .Fl.c.s1c2 */

int
process_request(slist *req) {
	struct url *curl;
	int status;

	memset(&cd, 0, sizeof(cd));

	/* Size */
	cd.In = atoi(tsize); /* Requested object size */

	/*
	** We won't account plain resource wasting
	*/

	if(((int)cd.In) <= 0)
		return 1;

	if(*treq == 'I')	/* ICP... method */
		return 1;

	status = get_status(tstatus);

	/* Invalid status code */
	if(status < 0)
		return -1;

	cd.TS = atol(tstamp);
	if(!inet_aton(tfromip, (struct in_addr *)&cd.Src)) {
		fprintf(stderr, "Failed to inet_aton()...\n");
		return -1;
	};

	/* Time spent */
	cd.Out = atoi(tspent);
	if( ((int)cd.Out) < 0 )
		cd.Out = 0;

	cd.FL.c.s1c1 = get_way(tway);

	curl = split_url(turl);
	if(!curl) {
		fprintf(stderr, "Can't parse url\n");
		return -1;
	};

	/* Get MIME-type */
	cd.FL.c.s1c2 = get_mime(tmime, curl->path);

	if(expressions)
		cd.FL.c.s2c1 = get_urlre( curl->url );
	cd.FL.c.s2c2 |= get_request_method(treq, curl->proto );

	cd.Dst = compute_domain_index( curl->domain, ((status==503)?0:1) );

	cd.U1=get_who(twho);

	ctb_add(ctbs, &cd);

	bytesctb += sizeof(cd);

	return 0;
};


char
get_way(char *way) {
	char r = 0;

	if(!way)
		return r;

	switch(*way) {
	case 'N': return 0;	/* NONE/NULL, cache hit */
	case 'D': return 1;	/* DIRECT */
	case 'P': return 2;	/* PROXY/PARENT */
	};

	return 2;
};


static char *mime_image[]={
	".gif", ".jpg", ".jpeg", ".tiff", ".bmp", ".pcx",
	NULL };

static char *mime_audio[]={
	"mp3", ".mp2", ".wav", ".voc", ".snd", ".mid",
	NULL };

static char *mime_archive[]={
	".zip", ".arj", ".rar", ".lzh", ".zoo", ".gzip", ".gz", ".rar", ".arc",
	".bs2", ".bz", ".bz2", ".ha", ".tgz",
	NULL };

char
get_mime(char *mime, char *path) {
	register char *p;

	/* Canonicalization. */
	for(p = mime; *p; p++) {
		if(*p >= 'A' && *p <= 'Z') {
			*p += 32;
		};
	};

	/* Text, directly displayable */
	if(strncmp(mime, "text/", 5) == 0) {
		if(eq(mime + 5, "plain")) return 0x01;
		if(eq(mime + 5, "html")) return 0x02;
		if(eq(mime + 5, "css")) return 0x03;
		return 0x00;
	};

	/* Images */
	if(strncmp(mime, "image/", 6) == 0) {
		if(eq(mime + 6, "gif")) return 0x11;
		if(eq(mime + 6, "jpeg")) return 0x12;
		return 0x10;
	};

	/* General binary */
	if(strncmp(mime, "application/", 12) == 0) {
		if(eq(mime + 12, "octet-stream")) {
			char *dot = NULL;

			if(!path) return 0x20;

			for(p = path; *p; p++) {
				if(*p == '.') {
					dot = p;
					continue;
				};
				if(*p == '/') {
					dot = NULL;
					continue;
				};
			};

			if(!dot) return 0x20;

			if(find(mime_archive, dot) != -1)
				return 0x21;
			if(find(mime_audio, dot) != -1)
				return 0x23;
			if(find(mime_image, dot) != -1)
				return 0x10;

			return 0x20;
		};

		/* Archives */
		if(eq(mime + 12, "zip")) return 0x21;
		if(eq(mime + 12, "x-tar")) return 0x21;
		if(eq(mime + 12, "x-gtar")) return 0x21;

		/* Documents */
		if(eq(mime + 12, "msword")) return 0x22;
		if(eq(mime + 12, "pdf")) return 0x22;
		if(eq(mime + 12, "rtf")) return 0x22;

		return 0x20;
	};

	/* Multimedia */
	if(strncmp(mime, "audio/", 6) == 0)
		return 0x23;
	if(strncmp(mime, "video/", 6) == 0)
		return 0x24;

	return 0x20;	/* As application/octet-stream */
};


unsigned int
get_who_passwd(char *user) {
	struct passwd *pw;

	if(!(pw=getpwnam(user)))
		return 0;

	return (unsigned int)pw->pw_uid;

};


unsigned int
get_who(char *user) {

	if(!user || !*user || (*user == '-'))
		return 0;

#ifdef	HAVE_LIBLGBA
	if(!usenode)
		return get_who_passwd(user);

	if(load(user))
		return 0;

	{
		uid_t uid;
		uid = atoi(getattr("uidnumber"));
		if(!uid)
			uid = atoi(getattr("uid"));

		return uid;
	}

#else
	return get_who_passwd(user);
#endif
};

char *
crop_domain(char *url, int level) {
	register char *endp;
	register int isnotnum;

	if(level <= 0)
		level = 255;

	/* Seek to the end */
	for(endp = url; *endp; endp++);

	isnotnum = 0;

	for(--endp; endp > url; endp--) {
		if(*endp == '.') {
			if(!--level && isnotnum)
				return (endp+1);
			continue;
		};
		if(*endp > '9' || *endp < '0')
			isnotnum = 1;
	};

/*
	if(!isnotnum) {
		endp = url;
		level = 3;
		for(endp = url; *endp; endp++) {
			if(*endp == '.')
				if(!--level) {
					*endp = '\0';
					return url;
				};
		};
	};
*/

	return url;
};


char
get_request_method(char *req, char *proto) {

	if(
		strcmp(req, "GET") == 0 ||
		strcmp(req, "POST") == 0 ||
		strcmp(req, "HEAD") == 0
	) {

		if(!proto) return 0x00;

		if(eq(proto, "http")) return 0x01;
		if(eq(proto, "ftp")) return 0x02;

		return 0x00;
	};

	if(strcmp(req, "CONNECT") == 0)
		return 0x03;

	return 0x00;
};

typedef struct _domain_cache_entry {
	char *domain;
	unsigned int index;
	struct _domain_cache_entry	*next;
} domain_cache_entry;

domain_cache_entry *dcache=NULL;

#define	MAX_DOMAIN_CACHE_SIZE	128

unsigned int
get_cached_domain_index(char *dmn) {
	register domain_cache_entry *dc;
	register domain_cache_entry *pdc; /* previous value*/
	register int count=0;

	dc = dcache;
	pdc = NULL;

	/* Try to find in cache */
	while(dc) {
		if(eq(dmn, dc->domain)) {
			/* Move this entry to the top */
			if(pdc) {
				/* Collapse previous and next */
				pdc->next = dc->next;
				dc->next = dcache;
				dcache = dc;
			};
			return dc->index;
		};
		pdc = dc;
		dc = dc->next;

		if(count++ > MAX_DOMAIN_CACHE_SIZE) {
			count = 0;
			pdc->next = NULL;

			/* Clear the last */
			while(dc) {
				pdc = dc->next;
				free(dc->domain);
				free(dc);
				dc=pdc;
				count++;
			};
			break;
		};
	};

	/* Cached index not found */
	return 0;

};


void
cache_domain_index(char *dmn, unsigned int idx) {
	register domain_cache_entry *dc;

	if(!dmn || !idx) return;

	if((dc=dcache)) {
		dcache=(domain_cache_entry *)
			malloc(sizeof(domain_cache_entry));
		if(!dcache) return;	/* Low in memory */
		dcache->next=dc;
		dc=dcache;
	} else {
		dc=dcache=(domain_cache_entry *)
			malloc(sizeof(domain_cache_entry));
		if(!dc) return; /* Low in memory */
		dc->next=NULL;
	};

	dc->domain = strdup(dmn);
	dc->index = idx;

	if(!dc->domain) {
		if(dc->next)
			dcache = dc->next;
		else
			dcache = NULL;
		free(dc);
	};

};

#define dhash(h, c) ((h) = 0x53*(h) + 0xac + (c))

unsigned char
hash_tld(char *domain) {
	unsigned char h = 0;

	for(; *domain; domain++)
		dhash(h, (unsigned char)*domain);

	return h;
};

unsigned int
compute_domain_index(char *domain, int add2map) {
	register unsigned int idx;

	if(!domain || !*domain)
		return 0;

	/* Is this domain already cached? */
	idx = get_cached_domain_index(domain);
	if(idx)
		return idx;

	/* Try to find index in the map */
	idx = ctb_revmap(ctbs, FIELD_DST, domain, RVM_EXACT, NULL, NULL);
	if(idx) {
		cache_domain_index(domain, idx);
		return idx;
	};

	if(!add2map)	/* We has the order not to add this to the map */
		return 0;

	/* Get the top level domain hash. */

	idx = ((unsigned int)hash_tld(crop_domain(domain, 1))) << 24;
	idx |= ((ctb_map_size(ctbs, FIELD_DST) + 1) & 0x00ffffff);

	if(ctb_addmap(ctbs, FIELD_DST, idx, domain) != 0)
		cache_domain_index(domain, idx);

	return idx;
};


int
get_status(char *st) {
	int scode;
	register char *s;

	if(*st == 'N' && strncmp(st, "NONE/", 5) == 0)
		return -1;

	for(s = st; *s; s++) {
		if(*s == '/')
			break;
	};
	if(!*s) return -1;
	*s++ = '\0';

	/* Ignore it! */
	if(strncmp(st, "TCP_", 4) == 0) {
		if(strcmp(st+4, "DENIED") == 0)
			return 0;
	};

	scode = strtoul(s, NULL, 10);

	return scode;
};



