#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_LIBCOST
#include <cost.h>
#endif

#include <strfunc.h>

#include "desire.h"
#include "ddraw.h"

static char mime_types[256];
int strict_mime_types = 0;

static char domain_hashes[256];
unsigned int *domains = NULL;
size_t domains_count = 0;
int domains_error = 0;

struct _whs {
	unsigned int what;
	unsigned int hits;
	long long size;
};

struct _whs *found_domains = NULL;
size_t found_domains_count = 0;

struct _whs *found_ips = NULL;
size_t found_ips_count = 0;

int comparH(const void *, const void *);
int comparS(const void *, const void *);

int add_domain_index(unsigned int, char const *, void *);
int add_found_domain(unsigned int Idx, size_t size);
int add_found_ip(unsigned int Idx, size_t size);

int main(int ac, char **av) {
	struct desire_image_options *io;
	ctb_data *data;
	int count;
	ctb_session *ctbs = NULL;
	int lockMyself;
	slist *sl;
	char *s;
	int top = 0;

	printf("Expires: %s\n", timetostr(time(NULL)+300,
		TFMT_RFC822 | TFMT_OLD822GMT));
	printf("Content-type: text/html\n\n");
	fflush(stdout);

	if( (lockMyself = open(av[0], O_RDONLY)) != -1)
		lockf(lockMyself, F_LOCK, 1);

	io = fill_desire_image_options(0);

	if((s=param("top")))
		top = atoi(s);

	if(!io->service || !(ctbs=ctb_open(NULL, io->service, io->file))) {
		data=NULL; count=0;
	} else {
		data=ctb_get(ctbs, &(io->range), &count);

		if(io->list1) {
			int i;
			for(i = 0; i < io->list1->count; i++)
				ctb_revmap(ctbs, FIELD_DST, io->list1->list[i], RVM_FUZZY, add_domain_index, io);
		};

		if((sl=param2("mime", 1))) {
			unsigned int t;
			int i, k;
			for(i = 0; i < sl->count; i++) {
				errno = 0;
				t = strtoul(sl->list[i], NULL, 0);
				if(errno || t > 255) continue;
				mime_types[t] = 1;
				strict_mime_types = 1;
				if(!(t & 0x0f))
					for(k = 0; k <= 0x0f; k++)
						mime_types[(t & 0xf0) + k] = 1;
			};
		};

		if(count > 0 && io->trim) {
			io->range.from = data->TS;
			io->range.to = (data+count-1)->TS;
		};
	};

	printf("<HTML><BODY BGCOLOR=white><CENTER>");

	printf("<IMG SRC=proxy_draw.cgi?%s><P>", getenv("QUERY_STRING"));

	if(count) {
		int dc;
		ctb_data *dt;
		int mix = io->mix2;
		unsigned int ip = io->ip;
		unsigned int mask = io->mask;
		int dat;
		int dmc;
		struct _whs *whs;
		unsigned int *domain;
		unsigned int dmn;

		for(dc = count, dt = data; dc; dt++, dc--) {

			if(mask)
				if((dt->Src & mask) != ip) 
					continue;

			if(mix > 0) {
				dat = dt->FL.c.s2c2 & 0x03;
				switch(dat) {
				case 0x01: if(!(mix & 1)) dt->TS = 0; break; /* HTTP */
				case 0x02: if(!(mix & 2)) dt->TS = 0; break; /* FTP */
				case 0x03: if(!(mix & 4)) dt->TS = 0; break; /* CONNECT */
				case 0x00: if(!(mix & 8)) dt->TS = 0; break; /* UNKNOWN */
				};
				if(!dt->TS)
					continue;
			};

			if(strict_mime_types) {
				if(!mime_types[(unsigned char)dt->FL.c.s1c2])
					continue;
			};

			dmn = dt->Dst;

			/* Check for specified domains */
			if(domains_count) {
				if(!domain_hashes[(unsigned char)(dmn >> 24)])
					continue;

				for(domain=domains, dmc=domains_count;
					dmc; dmc--, domain++) {

					if(*domain != dmn)
						continue;

					if((domain - domains) < 5)
						break;

					for(; domain > domains; domain--)
						*domain = domain[-1];

					*domains = dmn;

					dmc = 0;
					break;
				};

				if(!dmc)
					continue;
			};


			/* Found domains */
			for(whs = found_domains,
				dmc = found_domains_count;
					dmc; dmc--, whs++) {
				if((*whs).what != dmn)
					continue;
				(*whs).hits++;
				(*whs).size += dt->In;
				if((whs - found_domains) < 5)
					break;
				{
					static struct _whs tmp;
					tmp = *whs;
					for(; whs > found_domains; whs--)
						*whs = whs[-1];
					*found_domains = tmp;
				};
				break;
			};
			if(!dmc)
				add_found_domain(dmn, dt->In);

			dmn = dt->Src;
			/* Found IPS */
			for(whs = found_ips,
				dmc = found_ips_count;
					dmc; dmc--, whs++) {
				if((*whs).what != dmn)
					continue;
				(*whs).hits++;
				(*whs).size += dt->In;

				if((whs - found_ips) < 5)
					break;
				{
					static struct _whs tmp;
					tmp = *whs;
					for(; whs > found_ips; whs--)
						*whs = whs[-1];
					*found_ips= tmp;
				};
				break;
			};
			if(!dmc)
				add_found_ip(dmn, dt->In);

		};
	};

	fflush(stdout);

	if(!found_domains_count) {
		printf("<FONT SIZE+1>No domains found</FONT>");
	} else {
		int dmc;
		int dto = found_domains_count<top?found_domains_count:top;

		printf("<FONT SIZE+1>%u domains found</FONT><P>", found_domains_count);

		printf("<TABLE BORDER=0><TR><TD>");

		qsort(found_domains, found_domains_count,
			sizeof(*found_domains), comparS);
		printf("<TABLE BORDER=1><TR><TH>No.</TH><TD> Domain </TD> <TD> Hits </TD> <TD> Size </TD> </TR>\n");
		for(dmc = 0; dmc < dto; dmc++) {
			printf("<TR><TH> %d. </TH><TD> %s </TD><TD> %u </TD><TD> ",
				dmc+1,
				ctb_map(ctbs, &found_domains[dmc].what, FIELD_DST),
				found_domains[dmc].hits);

			if(found_domains[dmc].size < 5000) {
				printf("%qu", found_domains[dmc].size);
			} else if(found_domains[dmc].size < 1000000) {
				printf("%qu k", found_domains[dmc].size >> 10);
			} else {
				printf("%.1f M", (float)found_domains[dmc].size / (1024*1024));
			};

			printf(" </TD></TR>\n");
		};
		printf("</TABLE>\n");

		printf("</TD><TD>\n");

		qsort(found_domains, found_domains_count,
			sizeof(*found_domains), comparH);
		printf("<TABLE BORDER=1><TR><TH>No.</TH><TD> Domain </TD> <TD> Hits </TD> <TD> Size </TD> </TR>\n");
		for(dmc = 0; dmc < dto; dmc++) {
			printf("<TR><TH> %d. </TH><TD> %s </TD><TD> %u </TD><TD> ",
				dmc+1,
				ctb_map(ctbs, &found_domains[dmc].what, FIELD_DST),
				found_domains[dmc].hits
			);

			if(found_domains[dmc].size < 5000) {
				printf("%qu", found_domains[dmc].size);
			} else if(found_domains[dmc].size < 1000000) {
				printf("%qu k", found_domains[dmc].size >> 10);
			} else {
				printf("%.1f M", (float)found_domains[dmc].size / (1024*1024));
			};
			printf(" </TD></TR>\n");
		};
		printf("</TABLE>\n");

		printf("</TD></TABLE>\n");

	};

	printf("<P>\n");

	if(!found_ips_count) {
		printf("<FONT SIZE+1>No clients found</FONT>");
	} else {
		int ipc;
		int dto = found_ips_count<top?found_ips_count:top;

		printf("<FONT SIZE+1>%u clients found</FONT><P>", found_ips_count);

		printf("<TABLE BORDER=0><TR><TD>");

		qsort(found_ips, found_ips_count,
			sizeof(*found_ips), comparS);
		printf("<TABLE BORDER=1><TR><TH>No.</TH><TD> Domain </TD> <TD> Hits </TD> <TD> Size </TD> </TR>\n");
		for(ipc = 0; ipc < dto; ipc++) {
			printf("<TR><TH> %d. </TH><TD> %s </TD><TD> %u </TD><TD> ",
				ipc+1,
				inet_ntoa(*(struct in_addr *)&found_ips[ipc].what),
				found_ips[ipc].hits);

			if(found_ips[ipc].size < 5000) {
				printf("%qu", found_ips[ipc].size);
			} else if(found_ips[ipc].size < 1000000) {
				printf("%qu k", found_ips[ipc].size >> 10);
			} else {
				printf("%.1f M", (float)found_ips[ipc].size / (1024*1024));
			};

			printf(" </TD></TR>\n");
		};
		printf("</TABLE>\n");

		printf("</TD><TD>\n");

		qsort(found_ips, found_ips_count,
			sizeof(*found_ips), comparH);
		printf("<TABLE BORDER=1><TR><TH>No.</TH><TD> Domain </TD> <TD> Hits </TD> <TD> Size </TD> </TR>\n");
		for(ipc = 0; ipc < dto; ipc++) {
			printf("<TR><TH> %d. </TH><TD> %s </TD><TD> %u </TD><TD> ",
				ipc+1,
				inet_ntoa(*(struct in_addr *)&found_ips[ipc].what),
				found_ips[ipc].hits
			);

			if(found_ips[ipc].size < 5000) {
				printf("%qu", found_ips[ipc].size);
			} else if(found_ips[ipc].size < 1000000) {
				printf("%qu k", found_ips[ipc].size >> 10);
			} else {
				printf("%.1f M", (float)found_ips[ipc].size / (1024*1024));
			};
			printf(" </TD></TR>\n");
		};
		printf("</TABLE>\n");

		printf("</TD></TABLE>\n");

	};

	printf("</BODY></HTML>");

	ctb_close(ctbs);

	return 0;
};

int
add_domain_index(unsigned int Idx, char const *Value, void *optKey) {
	struct desire_image_options *io = (struct desire_image_options *)optKey;
	int *d;

	if(!io) return -1;

	d = realloc(domains, sizeof(*domains) * (++domains_count));
	if(!d) {
		domains_error = 1;
		domains_count = 0;
		return -1;
	};
	domains = d;

	domain_hashes[(unsigned char)(Idx >> 24)] = 1;

	domains[domains_count - 1] = Idx;

	return 0;
};

int
add_found_domain(unsigned int Idx, size_t size) {
	found_domains_count++;
	found_domains = sf_realloc(found_domains,
		sizeof(struct _whs) * found_domains_count);
	found_domains[found_domains_count - 1].what = Idx;
	found_domains[found_domains_count - 1].hits = 1;
	found_domains[found_domains_count - 1].size = size;
	return 0;
};

int
add_found_ip(unsigned int Idx, size_t size) {
	found_ips_count++;
	found_ips = sf_realloc(found_ips,
		sizeof(struct _whs) * found_ips_count);
	found_ips[found_ips_count - 1].what = Idx;
	found_ips[found_ips_count - 1].hits = 1;
	found_ips[found_ips_count - 1].size = size;
	return 0;
};

int comparH(const void *ap, const void *bp) {
	register struct _whs *a = (struct _whs *)ap;
	register struct _whs *b = (struct _whs *)bp;
	return b->hits - a->hits;
};

int comparS(const void *ap, const void *bp) {
	register struct _whs *a = (struct _whs *)ap;
	register struct _whs *b = (struct _whs *)bp;
	return b->size - a->size;
};
