#ifndef	__DDRAW_H__
#define	__DDRAW_H__

#include <gd.h>

#include <desire.h>

#ifndef	gdFontSmall
extern gdFontPtr gdFontSmall;
#endif

#ifndef	gdFontTiny
extern gdFontPtr gdFontTiny;
#endif


/*
** Line filter structure.
**
** Line filters used to select data chunks which are to be counted
** when displaying an appropriate line by ddraw().
**
*/

struct line_filter {

	/* Use only optFilter if Offset too high */
	unsigned int Offset;  /* That byte ... */
	unsigned int Mask;    /* ... should be intersected with this mask ... */
	unsigned int Value;   /* .. and form this value. */
	unsigned int Rule;    /* Accept (((data+Offset) & Mask) == Value)? */

	/* Clears FIELD_TS to exclude (filter) */
	void (*optFilter)(ctb_data *data);
};

struct custom_line {

	/* Color values will contain color indexes upon return from ddraw() */
	int foreColor;   /* 0x00rrggbb */
	int fillColor;   /* 0x00rrggbb, if -1 then it will be ignored */

	/* Define meaningful value (_x_ bytes received) */
	unsigned int dataOffset;  /* Value offset, FIELD_xx (i.e. FIELD_IN) */
	unsigned int dataMask;    /* Mask to apply */

	/* Define meaningful value width (x bytes received in _y_ seconds) */
	unsigned int widthOffset; /* FIELD_xx, FIELD_OUT, in seconds. */
	unsigned int widthMask;   /* Mask to apply */
	int widthSign;	/* -1: TS is the start time, 1: TS is the end time */

	/* Optional line filter */
	struct line_filter *optLineFilter;	/* optional Filter */

	/* Data will be normalized according to this group */
	int group;

	/*************************************/
	/* ddraw() will fill the rest itself */
	/*************************************/

	long long *points;
	size_t	*stack;

	unsigned long meaningful_points;

	float factor;   /* multiply factor to shrink into allowed area */

	unsigned int elmax;  /* elementary maximum of data chunk */
	unsigned int stackmax;	/* gathered stack maximum */

	long long max;  /* gathered maximum of pixel value */

	long long all;  /* gathered All */
};

struct desire_image_options {

	/* DDF_TIME_RANGE */
	ctb_range range;	/* Time range */

	/* DDF_WIDTH */
	int image_width;	/* Desired width */

	/* DDF_HEIGHT */
	int image_height;	/* Desired height */

	/* DDF_AREA_HEIGHT */
	int area_height;

	/* DDF_LEFT_OFFSET */
	int left_offset;

	/* DDF_VERTICAL_LEGEND */
	char *vertical_legend;

	/* DDF_LEGEND_FACTOR */
	unsigned int legend_factor;

	/* DDF_SERVICE */
	char *service;	/* CTB service */

	/* DDF_FILE */
	char *file;		/* CTB subname */

	/* DDF_SMOOTH */
	int smooth;	/* Graph smoothing */

	/* DDF_TRIM */
	int trim;	/* Left and right trim */

	/* DDF_LOGARITHM */
	int logarithm;	/* Logarithmic normalization */

	/* DDF_INTERLACE */
	int interlace;	/* Image interlace option */

	/* DDF_COMMENTS */
	slist *comments;	/* Custom image commentaries */

	/* Less common options */

	/* For LIBCOST (-lcost) to show cost of data stream */
	/* DDF_COST_OPTIONS */
	char *costtype;
	char *costclass;

	/* DDF_IPM */
	unsigned int ip;	/* Recognized IP */
	unsigned int mask;	/* And mask */

	/*******************/
	/* Unsorted fields */
	/*******************/

	/* DDF_LISTS */
	slist *list1;
	slist *list2;
	slist *list3;

	/* DDF_MIXES */
	int	mix1;	/* OR'ed value, or -1 */
	int	mix2;	/* OR'ed value, or -1 */
	int	mix3;	/* OR'ed value, or -1 */


	/**********************/
	/* Pre-filter engines */
	/**********************/

	struct line_filter *optPreFilter;	/* optional Pre-filter */
	void (*optBlockPreFilter)(
		ctb_session *ctbs,
		ctb_data *data,
		int dataCount,
		struct desire_image_options *io
		);

	/************************************/
	/* ddraw() will not use this at all */
	/************************************/

	void *optKey;

	/****************************************/
	/* ddraw() will fill the rest by itself */
	/****************************************/

	int inpixel;	/* Seconds in one pixel */
};


#define	DDF_TIME_RANGE       0x000001
#define	DDF_WIDTH            0x000002
#define	DDF_HEIGHT           0x000004
#define	DDF_AREA_HEIGHT      0x000008
#define	DDF_LEFT_OFFSET      0x000010
#define	DDF_VERTICAL_LEGEND  0x000020
#define	DDF_SERVICE          0x000040
#define	DDF_FILE             0x000080
#define	DDF_SMOOTH           0x000100
#define	DDF_TRIM             0x000200
#define	DDF_LOGARITHM        0x000400
#define	DDF_INTERLACE        0x000800
#define	DDF_COMMENTS         0x001000
#define	DDF_COST_OPTIONS     0x002000
#define	DDF_IPM              0x004000
#define	DDF_LEGEND_FACTOR    0x008000
#define	DDF_LISTS            0x010000
#define	DDF_MIXES            0x020000

struct desire_image_options *
	fill_desire_image_options(unsigned int optDDFOptionSet);

gdImagePtr
ddraw(
	ctb_session *ctbs,         /* DESIRE session pointer */
	  ctb_data *data,          /* Data vector */
	  int dataCount,           /* Data count */
	struct custom_line *line,  /* Define lines */
	int lines,                 /* Lines quantity */
	struct desire_image_options *io /* Image options */
);

#endif
