#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>

#include <strfunc.h>

#include "desire.h"
#include "ddraw.h"

struct desire_image_options *
fill_desire_image_options(unsigned int ddf) {
	static struct desire_image_options dio;
	char *s;
	slist *sl;

	char *d, *m, *y;
	char *d2, *m2, *y2;
	struct tm tm, tm2;

	if(!ddf)
		ddf = (unsigned int)-1;

	memset(&dio, 0, sizeof(dio));

	if((ddf & DDF_SERVICE))
		dio.service=param("service");

	if((ddf & DDF_FILE))
		dio.file=param("file");

	if((ddf & DDF_TRIM)) {
		dio.trim=param("trim")?1:0;
		if(dio.file && (strchr(dio.file, '/') || strstr(dio.file, ".."))) {
			dio.service=NULL; dio.file=NULL;
		}
	}

	if((ddf & DDF_LOGARITHM))
		dio.logarithm=param("log")?1:0;

	if((ddf & DDF_COMMENTS))
		dio.comments=scopy(param2("comments", 0));

	if((ddf & DDF_VERTICAL_LEGEND)) {
		dio.vertical_legend = param("vertical_legend");
		if(dio.vertical_legend)
			dio.vertical_legend = sf_strdup(dio.vertical_legend);
	}
	
	if((ddf & DDF_INTERLACE))
		if((s=param("interlace")) && strcmp(s, "off"))
			dio.interlace = 1;

	if((ddf & DDF_COST_OPTIONS)) {
		dio.costtype=param("costtype");
		if(dio.costtype && !*dio.costtype)
			dio.costtype = 0;
		dio.costclass=param("costclass");
	}

	/*********/
	/* Mixes */
	/*********/

  if(ddf & DDF_MIXES) {

	dio.mix1 = -1;
	dio.mix2 = -1;
	dio.mix3 = -1;

	if((sl = param2("mix1", 0))) {
		int i;
		dio.mix1 = 0;
		for(i = 0; i < sl->count; i++) {
			s = sl->list[i];
			dio.mix1 |= strtoul(s, NULL, 0);
		};
	};

	if((sl = param2("mix2", 0))) {
		int i;
		dio.mix2 = 0;
		for(i = 0; i < sl->count; i++) {
			s = sl->list[i];
			dio.mix2 |= strtoul(s, NULL, 0);
		};
	};

	if((sl = param2("mix3", 0))) {
		int i;
		dio.mix3 = 0;
		for(i = 0; i < sl->count; i++) {
			s = sl->list[i];
			dio.mix3 |= strtoul(s, NULL, 0);
		};
	};
  }

	if((ddf & DDF_WIDTH)) {
	  dio.image_width = 620;
	  if((s=param("width")))
		dio.image_width = atoi(s);
	}

	if((ddf & DDF_HEIGHT)) {
	  dio.image_height = 160;
	  if((s=param("height"))) {
		dio.image_height = atoi(s);
		if(dio.image_height < 50)
			dio.image_height = 50;
	  }
	}

	if((ddf & DDF_AREA_HEIGHT)) {
	  dio.area_height = 100;
	  if((s=param("area_height"))) {
		dio.area_height = atoi(s);
		if(dio.area_height < 50)
			dio.area_height = 50;
	  }
	}

	if((ddf & DDF_LEFT_OFFSET)) {
	  dio.left_offset = 100;
	  if((s=param("left_offset"))) {
		dio.left_offset = atoi(s);
		if(dio.left_offset < 50)
			dio.left_offset = 50;
	  }
	}

	if((ddf & DDF_LEGEND_FACTOR)) {
	  dio.legend_factor = 0;
	  if((s=param("legend_factor"))) {
		dio.legend_factor = atoi(s);
	  }
	}

	if((ddf & DDF_SMOOTH)) {
	  dio.smooth=1;
	  if((s=param("smooth"))) {
		dio.smooth=atoi(s);
		if(dio.smooth < 1)  dio.smooth = 1;
		if(dio.smooth > 50) dio.smooth = 50;
	  }
	}

	if((ddf & DDF_IPM) && (s = param("ip"))) {
		if(split_network(s, &dio.ip, &dio.mask))
			dio.ip = dio.mask = 0;
	};

	/*********/
	/* Lists */
	/*********/

  if(ddf & DDF_LISTS) {

	if((s = param("list1"))) {
		dio.list1 = split(s, ", ", 0);
		if(!dio.list1->count) {
			sfree(dio.list1);
			dio.list1 = NULL;
		}
	}

	if((s = param("list2"))) {
		dio.list2 = split(s, ", ", 0);
		if(!dio.list2->count) {
			sfree(dio.list2);
			dio.list2 = NULL;
		}
	}

	if((s = param("list3"))) {
		dio.list3 = split(s, ", ", 0);
		if(!dio.list3->count) {
			sfree(dio.list3);
			dio.list3 = NULL;
		}
	}

  }



	/***************/
	/* Time ranges */
	/***************/

  if((ddf & DDF_TIME_RANGE)) {

	dio.range.either = 1;
	dio.range.simple_def = LAST | DAY;

	if((s=param("dmy"))) {
		dio.range.either=1;
		dio.range.simple_def &= 0xf0;
		switch(*s) {
		case 'h':
			dio.range.simple_def |= HOUR;
			break;
		case 'd':
			dio.range.simple_def |= DAY;
			break;
		case 'w':
			dio.range.simple_def|=WEEK;
			break;
		case 'm': dio.range.simple_def|=MONTH; break;
		case 'y':
			dio.range.simple_def|=YEAR;
			break;
		};
	};

	if((s=param("clp"))) {
		dio.range.either=1;
		dio.range.simple_def &= 0x0f;
		switch(*s) {
			case 'c': dio.range.simple_def |= CURR; break;
			case 'l': dio.range.simple_def |= LAST; break;
			case 'p': dio.range.simple_def |= PREV; break;
		};
	};

	ctb_fill_time(&dio.range, dio.range.simple_def);

	if((s=param("manual")) && !strcmp(s, "off")) return &dio;

	y=param("y");
	m=param("m");
	d=param("d");
	d2=param("d2");
	m2=param("m2");
	y2=param("y2");

	if(!(y || m || d || d2 || m2 || y2)) return &dio;

	tm.tm_sec=tm.tm_min=tm.tm_hour=0;
	tm2.tm_sec=tm2.tm_min=tm2.tm_hour=0;

	if(!(dio.range.from = strtotime(param("1"))))  {

		if(y && atoi(y) > 0) {
			tm.tm_mday=1; tm.tm_mon=0; tm.tm_isdst=0;
			tm.tm_year=atoi(y);
			tm2.tm_mday=1; tm2.tm_mon=0; tm2.tm_isdst=0;
			tm2.tm_year=atoi(y)+1;
		};
	
		if(m && atoi(m) > 0) {
			tm.tm_mday=1;
			tm.tm_mon=atoi(m)-1;
			tm2.tm_mday=1;
			tm2.tm_year=tm.tm_year;
			tm2.tm_mon=atoi(m);
		};
	
		if(d && atoi(d) > 0) {
			tm.tm_mday=atoi(d);
			tm2.tm_mday=tm.tm_mday+1;
			tm2.tm_year=tm.tm_year;
			tm2.tm_mon=tm.tm_mon;
		};

	}; /* if(!(dio.range.from = strototime(param("1"))))  */


	if(!(dio.range.to = strtotime(param("2")))) {

		if(y2 && atoi(y2) > 0) {
			tm2.tm_mday=1;
			tm2.tm_mon=0;
			tm2.tm_year=atoi(y2);
		};
		if(m2 && atoi(m2) > 0) {
			tm2.tm_mday=1;
			tm2.tm_mon=atoi(m2)-1;
		};
		if(d2 && atoi(d2) > 0) {
			tm2.tm_mday=atoi(d2);
		};

	};

	if(!dio.range.from) {
		if(tm.tm_year > 1900) tm.tm_year-=1900;
		dio.range.from=mktime(&tm);
		if(tm.tm_isdst)
			dio.range.from -= 3600;
	};

	if(!dio.range.to) {
		if(tm2.tm_year > 1900) tm2.tm_year-=1900;
		dio.range.to=mktime(&tm2);
		if(tm2.tm_isdst)
			dio.range.to -= 3600;
	};

	dio.range.either=0;
  }

	return &dio;
};


