%{

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fnmatch.h>

#include "desire.h"
#include "session.h"
#include "ctbint.h"

int _ctb_lineno;

int yylex();
int yyerror();

#define	YYPARSE_PARAM	ctbs
%}

/*
 * Token value definition
 */
%union {
	char *tv_char;
	int  tv_int;
	ctb_sess *tv_sess;
	twooffsets tv_offs;
}

/*
 * Token types returned by scanner
 */
%token <tv_char>	TOK_STRING
%token <tv_int>		TOK_INTEGER
%token				ERROR

%token				EQ
%token				LB
%token				RB
%token				COMMA
%token				COLON
%token				SEMICOLON
%token				C_SERVICE
%token				C_FOLDER
%token				C_MAP
%token				C_FL
%token				C_IN
%token				C_OUT
%token				C_SRC
%token				C_DST
%token				C_U1
%token				C_U2
%token				C_DESCR
%token				C_ENTRYSIZE

%type	<tv_offs>	Field
%type	<tv_sess>	Session
%type	<tv_sess>	Session_def

%%
cfg_file:
	| sequence {
		if(((ctb_sess *)ctbs)->tmp_format_string) {
			free(((ctb_sess *)ctbs)->tmp_format_string);
			((ctb_sess *)ctbs)->tmp_format_string = NULL;
		};
	};

sequence: block ps
	| sequence block ps
	;

ps:
	| SEMICOLON
	;

block:
	C_SERVICE TOK_STRING LB Session ps RB {
		ctb_sess *session=$4;
		_ctb_map *map;

		if(strcmp($2, ((ctb_sess *)ctbs)->service) != 0) {
			ctb_close(session);
			break;
		};

		/* Playing with entry size */
		if(session->EntrySize == 0)
			session->EntrySize = ((ctb_sess *)ctbs)->EntrySize;
		if(session->EntrySize < sizeof(time_t)) {
			fprintf(stderr, "Invalid data entry size: %d.\n",
				session->EntrySize);
			return -1;
		};
		((ctb_sess *)ctbs)->EntrySize = session->EntrySize;

		if(!session->folder) {
			fprintf(stderr,
				"Folder or subfolder must be defined for session %s\n", $2);
			return -1;
		};

		if(((ctb_sess *)ctbs)->folder)
			free(((ctb_sess *)ctbs)->folder);
		((ctb_sess *)ctbs)->folder=session->folder;
			session->folder=NULL;
		((ctb_sess *)ctbs)->description=session->description;
			session->description = NULL;
		((ctb_sess *)ctbs)->maps = session->maps;
		session->maps = NULL;

		for(map = ((ctb_sess *)ctbs)->maps; map; map=map->next) {
			if(map->off_len < 1 || map->off_len > 4) {
				fprintf(stderr, "Invalid mapping lengh range: %d.\n",
					map->off_len);
				return -1;
			};
			if(map->off_start < sizeof(time_t)
			|| (map->off_start + map->off_len) > ((ctb_sess *)ctbs)->EntrySize) {
				fprintf(stderr, "Invalid mapping: %d bytes from %d is not within %d bytes.\n", map->off_len, map->off_start, ((ctb_sess *)ctbs)->EntrySize);
				return -1;
			};
		};

		ctb_close(session);

		((ctb_sess *)ctbs)->ready = 1;
		return 0;
	}
	| C_FOLDER EQ TOK_STRING {
		if(((ctb_sess *)ctbs)->folder)
			free(((ctb_sess *)ctbs)->folder);
		((ctb_sess *)ctbs)->folder = $3;
		if(*((ctb_sess *)ctbs)->folder != '/') {
			fprintf(stderr, "Global folder definition must begin with '/'.\n");
			return -1;
		};
	}
	;

Session:
	Session_def {
		$$ = $1;
	}
	| Session ps Session_def {
		_ctb_map *mto, *mfrom;

		$$=$1;

		if($3->description) {
			if($$->description)
				free($$->description);
			$$->description = $3->description;
			$3->description = NULL;
		};

		for(mfrom=$3->maps; mfrom; mfrom=mfrom->next) {
			for(mto = $$->maps; mto; mto=mto->next) {
				if((mfrom->off_start >= mto->off_start
				&& mfrom->off_start < (mto->off_start + mto->off_len))
				|| (mto->off_start >= mfrom->off_start
				&& mto->off_start < (mfrom->off_start + mfrom->off_len))) {
					fprintf(stderr, "Mapping overlaps near line %d.\n",
						_ctb_lineno);
					return -1;
				};
			};
			if((mto = $$->maps)) {
				while(mto->next) mto=mto->next;
				mto=mto->next=(_ctb_map *)_ctb_malloc(sizeof(_ctb_map));
			} else {
				mto = $$->maps = (_ctb_map *)_ctb_malloc(sizeof(_ctb_map));
			};
			memcpy(mto, mfrom, sizeof(_ctb_map));
			mto->next = NULL;
		};

		/* Clear maps */
		for(mfrom=mto=$3->maps; mfrom; mfrom=mto) {
			mto = mfrom->next;
			free(mfrom);
		};
		$3->maps = NULL;

		if($3->folder) {
			if($$->folder)	/* Mistake... */
				free($$->folder);
			$$->folder = $3->folder;
			$3->folder = NULL;
		};

		if($3->EntrySize) {
			$$->EntrySize = $3->EntrySize;
			$3->EntrySize = 0;
		};
	};

Session_def:
	C_FOLDER EQ TOK_STRING {
		ctb_sess *session;
		session = (ctb_sess *)_ctb_calloc(1, sizeof(ctb_sess));

		if(*($3) == '/') {
			session->folder=$3;
		} else {
			if(!((ctb_sess *)ctbs)->folder) {
				fprintf(stderr,
					"Global folder must be specified prior using relatives.\n");
				return -1;
			};
			session->folder = (char *)_ctb_malloc(
				strlen(((ctb_sess *)ctbs)->folder) + strlen($3) + 2);
			sprintf(session->folder, "%s/%s", ((ctb_sess *)ctbs)->folder, $3);
			free($3);
		};
		$$ = session;
	}
	| C_MAP Field EQ TOK_STRING {
		_ctb_map *map;
		$$ = (ctb_sess *)_ctb_calloc(1, sizeof(ctb_sess));
		map=$$->maps=(_ctb_map *)_ctb_calloc(1, sizeof(_ctb_map));
		map->mapfile = $4;
		map->off_start = $2.off_start;
		map->off_len = $2.off_len;
	}
	| C_DESCR EQ TOK_STRING {
		$$ = (ctb_sess *)_ctb_calloc(1, sizeof(ctb_sess));
		$$->description=$3;
	}
	| C_DESCR TOK_STRING EQ TOK_STRING {
		$$ = (ctb_sess *)_ctb_calloc(1, sizeof(ctb_sess));

		if(!((ctb_sess *)ctbs)->name || fnmatch($2, ((ctb_sess *)ctbs)->name, 0)) {
			free($2); free($4); break;
		};
		free($2);

		$$->description=$4;
	}
	| C_ENTRYSIZE EQ TOK_INTEGER {
		$$ = (ctb_sess *)_ctb_calloc(1, sizeof(ctb_sess));
		$$->EntrySize = $3;
	}
	;

Field:
	C_FL {
		$$.off_start=FIELD_FL;
		$$.off_len=sizeof(unsigned int);
	}
	| C_IN {	
		$$.off_start=FIELD_IN;
		$$.off_len=sizeof(unsigned int);
	}
	| C_OUT {	
		$$.off_start=FIELD_OUT;
		$$.off_len=sizeof(unsigned int);
	}
	| C_SRC {	
		$$.off_start=FIELD_SRC;
		$$.off_len=sizeof(unsigned int);
	}
	| C_DST {
		$$.off_start=FIELD_DST;
		$$.off_len=sizeof(unsigned int);
	}
	| C_U1 {
		$$.off_start=FIELD_U1;
		$$.off_len=sizeof(unsigned int);
	}
	| C_U2 {
		$$.off_start=FIELD_U2;
		$$.off_len=sizeof(unsigned int);
	}
	| TOK_INTEGER COLON TOK_INTEGER {
		$$.off_start=$1;
		$$.off_len=$3;
	}
	;

%%

int
yyerror(char *s) {
	fprintf(stderr, "Config parse error near line %d: %s\n", _ctb_lineno, s);
	return 0;
};
