%{

#include "session.h"
#include "ctbint.h"
#include "ctbconf.h"

int _ctb_lineno = 1;

%}

%option never-interactive 8bit
%option noyywrap stack
%option caseless pointer
%option noinput nounput

%pointer


%x comment
%x map
%x integer
%x value
%x quoted

%%

<*>"="	{ while(YY_START) yy_pop_state(); yy_push_state(value); return EQ; }
<*>";"	{ return SEMICOLON; }

#[^\n]*[\n]?	{ if(yytext[yyleng-1] == '\n') _ctb_lineno++; }

<*>"/*"	BEGIN(comment);
<comment>[^*/\n]	/* Eat */
<comment>"*"+[^*/\n]*	/* Eat */
<comment>\n	_ctb_lineno++;
<comment>"*"+"/"	BEGIN(INITIAL);


folder		{ return C_FOLDER; }
service	{
		yy_push_state(value);
		return C_SERVICE;
	}

"{"	return LB;
"}"	return RB;

map	{ return C_MAP; }
entry[a-z]*	{ return C_ENTRYSIZE; }
descr[a-z]*	{ yy_push_state(value); return C_DESCR; }


"FL"	{ return C_MAP; };
"In"	{ return C_IN; };
"Out"	{ return C_OUT; };
"Src"	{ return C_SRC; };
"Dst"	{ return C_DST; };
"U1"	{ return C_U1; };
"U2"	{ return C_U2; };
":"		{ return COLON; };
[0-9]+	{ _ctb_lval.tv_int = atoi(yytext); return TOK_INTEGER; };

<value>{
	[0-9]+	{
		yy_pop_state();
		_ctb_lval.tv_int = atoi(yytext);
		return TOK_INTEGER; }
	[a-z0-9/.-]+	{
		yy_pop_state();
		_ctb_lval.tv_char = _ctb_strdup(yytext);
		return TOK_STRING; }

	"\""	{
			yy_push_state(quoted);
		}
}

<quoted>{
	[^"]+	{ yymore(); }

	"\""	{
			yy_pop_state();
			if(YY_START == value)
				yy_pop_state();
			yytext[yyleng - 1] = 0;
			_ctb_lval.tv_char = _ctb_strdup(yytext);
			return TOK_STRING;
		}
}

<*>[[:space:]]	{ if(*yytext == '\n') _ctb_lineno++; }

<*>.	return ERROR;

%%
