#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "desire.h"
#include "session.h"

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef	HAVE_INTTYPES_H
#include <inttypes.h>
#endif

#ifdef	HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef	HAVE_FNMATCH_H
#include <fnmatch.h>
#endif

_ctb_map *
_ctb_findmap(ctb_sess *ctbs, int fld) {
	_ctb_map *map;

	if(!ctbs || !(map = ctbs->maps))
		return NULL;

	for(; map; map=map->next) {
		if(map->off_start == fld)
			return map;
	};

	return NULL;
};

DBMAP *
_ctb_map_open(char *mapfile, int rw) {
	DBMAP *dbf;
	int flags;

	if(rw) {
		flags = O_CREAT | O_RDWR
#ifdef	O_EXLOCK
			| O_EXLOCK
#endif
		;
	} else {
		flags = O_RDONLY
/*
#ifdef	O_SHLOCK
		| O_SHLOCK
#endif
*/
		;
	}

#ifdef	HAVE_DBOPEN
	dbf = (DBMAP *)dbopen(mapfile, flags, 0644, DB_BTREE, NULL);
#else
#ifdef	HAVE_DBM_OPEN
	dbf = (DBMAP *)dbm_open(mapfile, flags, 0644);
#endif
#endif

	if(!dbf)
		return (DBMAP *)0;

	if(!rw)
		return dbf;

#ifndef	O_EXLOCK
#ifdef  HAVE_FCNTL
	{
		struct flock fl = { 0, 0, 0, F_WRLCK, SEEK_SET };
		int fd;
		fl.l_pid = getpid();
		if(!rw)
			fl.l_type = F_RDLCK;

#ifdef	HAVE_DBOPEN
		fd = dbf->fd(dbf);
#else
#ifdef	HAVE_DBM_OPEN
		fd = dbm_dirfno(dbf);
#endif
#endif

		fcntl(fd, F_SETLKW, fl);
	}
#endif  /* FCNTL */
#endif

	return dbf;
}

int
_ctb_map_get(DBMAP *dbf, DBCHUNK *key, DBCHUNK *data) {

#ifdef	HAVE_DBOPEN
	return dbf->get(dbf, key, data, 0);
#else
#ifdef	HAVE_DBM_OPEN
	dbm_clearerr(dbf);
	*data = dbm_fetch(dbf, *key);
	if((data->map_data) && !dbm_error(dbf))
		return 0;
	else
		return -1;
#endif
#endif

};

int
_ctb_map_del(DBMAP *dbf, DBCHUNK *key) {

#ifdef	HAVE_DBOPEN
	return dbf->del(dbf, key, 0);
#else
#ifdef	HAVE_DBM_OPEN
	return dbm_delete(dbf, *key);
#endif
#endif

};

int
_ctb_map_put(DBMAP *dbf, DBCHUNK *key, DBCHUNK *data) {

#ifdef	HAVE_DBOPEN
	return dbf->put(dbf, key, data, 0);
#else
#ifdef	HAVE_DBM_OPEN
	{
		int rc;
		dbm_clearerr(dbf);
		rc = dbm_store(dbf, *key, *data, DBM_REPLACE);
		if(dbm_error(dbf))
			rc = -1;
		if(rc >= 1)
			rc = 0;
		return rc;
	}
#endif
#endif

};

void
_ctb_map_close(DBMAP *dbf) {

	if(!dbf)
		return;

#ifdef	HAVE_DBOPEN
	dbf->close(dbf);
#else
#ifdef	HAVE_DBM_OPEN
	dbm_close(dbf);
#endif
#endif

};



/*****************************/
/*** DESIRE map functions ****/
/*****************************/

int
ctb_addmap(ctb_session *ctbsp, int fld, unsigned int optValue, const char *str) {
	static char vl[sizeof(int) * 2];
	DBCHUNK av;
	DBCHUNK da;
	register DBMAP *dbf;
	register int rc;
	register _ctb_map *map;

	if(!(map=_ctb_findmap(ctbsp, fld)))
		return 0;

	if(!(dbf=map->dbf) || map->OpenedReadOnly) {
		if(dbf)
			_ctb_map_close(dbf);

		dbf = map->dbf = _ctb_map_open(map->mapfile, 1);

		map->OpenedReadOnly = 0;
		map->nentries = -1;
		map->lastindex = -1;

		if(!dbf)
			return 0;
	};

	if(str && !optValue) {
		optValue = ctb_map_lastno(ctbsp, fld);
		optValue++;
	}

	/* Removing platform dependency. */
	vl[sizeof(int) - 1] = 0;	/* Add leading zero */
	*(unsigned int *)(vl + sizeof(int)) = htonl(optValue);
	av.map_data = (char *)vl + sizeof(int) - 1;
	av.map_size = sizeof(int) + 1;

	if(str == NULL) {
		if (_ctb_map_get(dbf, &av, &da) == 0)
			_ctb_map_del(dbf, &da);
		if(_ctb_map_del(dbf, &av))
			return 0;
		return optValue;
	};

	da.map_data = (char *)str;
	da.map_size = strlen(str)+1;

	rc = _ctb_map_put(dbf, &av, &da);
	if(rc)
		return 0;

	av.map_data=(char *)str;
	av.map_size=strlen(str)+1;
	da.map_data = (char *)vl + sizeof(int) - 1;
	da.map_size = sizeof(int) + 1;

	if((rc = _ctb_map_put(dbf, &av, &da))) {
		/* Remove chunk. */
		ctb_addmap(ctbsp, fld, optValue, NULL);
		return 0;
	};

	if(map->nentries != -1)
		map->nentries++;

	if(optValue > map->lastindex)
		map->lastindex = optValue;

	return optValue;
};

char *
ctb_map(ctb_session *ctbsp, void *data, int fld) {
	char vl[sizeof(int) * 2];
	DBMAP *dbf;
	DBCHUNK av;
	DBCHUNK da;
	static char smb[16];
	unsigned int value = 0;
	_ctb_map *map;
	int i;
	char *c;

	if(!(map=_ctb_findmap(ctbsp, fld)))
		return NULL;	/* Mapping not found */

#ifdef	WORDS_BIGENDIAN
	for(i = 0, c=((char *)&value)+sizeof(value)-1; i < map->off_len; i++, c--) {
#else
	for(i = 0, c = (char *)&value; i < map->off_len; i++, c++) {
#endif
		*c = *(char *)((char *)data + i);
	};

	if(!(dbf=map->dbf)) {
		errno = 0;
		dbf = map->dbf = _ctb_map_open(map->mapfile, 0);

		if(!dbf) {
			sprintf(smb, "%u", value);
			return smb;
		};
	};

	vl[sizeof(int) - 1] = 0;
	*(unsigned int *)(vl + sizeof(int)) = htonl(value);
	av.map_data = (char *)vl + sizeof(int) - 1;
	av.map_size = sizeof(int) + 1;

	if(_ctb_map_get(dbf, &av, &da)) {
		sprintf(smb, "%u?", value);
		return smb;
	};

	return (char *)da.map_data;

};

unsigned int
ctb_revmap(
	ctb_session *ctbsp,
	int fld,
	const char *str,
	int flags,
	int (*func)(unsigned int, char const *, void *),
	void *optArg
) {
	DBMAP *dbf;
	DBCHUNK av;
	DBCHUNK da;
	unsigned int value=0;
	unsigned int values=0;
	int getall;
	_ctb_map *map;

	if(!str) return 0;

	if(!(map=_ctb_findmap(ctbsp, fld)))
		return 0;

	if(!(dbf=map->dbf)) {
		dbf = map->dbf = _ctb_map_open(map->mapfile, 0);

		map->OpenedReadOnly = 1;
		map->nentries = -1;
		map->lastindex = -1;

		if(!dbf)
			return 0;
	};

	if((flags & RVM_FUZZY) == 0) {
		av.map_data = (void *)str;
		av.map_size = strlen(str) + 1;
		if(_ctb_map_get(dbf, &av, &da))
			return 0;
		memcpy(&value, da.map_data + 1, da.map_size - 1);
		value = ntohl(value);
		return value;
	};

	if(!strcmp(str, "*"))
		getall = 1;
	else
		getall = 0;

#ifdef	HAVE_DBOPEN
	if(dbf->seq(dbf, &av, &da, R_FIRST) != 0)
		return 0;
#else
#ifdef	HAVE_DBM_OPEN
		dbm_clearerr(dbf);
	av = dbm_firstkey(dbf);
		if(!(av.map_data) || dbm_error(dbf)) return 0;
	da = dbm_fetch(dbf, av);
		if(!(da.map_data) || dbm_error(dbf)) return 0;
#endif
#endif

	for(;;) {
		if(*(char *)(av.map_data) != 0)	/* Not number, but string */
			goto crfc;

		if(!getall && fnmatch(str, da.map_data,
#ifdef	FNM_CASEFOLD
			FNM_CASEFOLD
#else
			0
#endif
			))
			goto crfc;

		memcpy(&value, av.map_data + 1, av.map_size - 1);
		value = ntohl(value);
		if(!func)
			return value;

		values++;

		if( func(value, (char const *)da.map_data, optArg) )
			return values;

	crfc:

#ifdef	HAVE_DBOPEN
		if(dbf->seq(dbf, &av, &da, R_NEXT) != 0)
			break;
#else
#ifdef	HAVE_DBM_OPEN
			dbm_clearerr(dbf);
		av = dbm_nextkey(dbf);
			if(!(av.map_data) || dbm_error(dbf)) break;
		da = dbm_fetch(dbf, av);
			if(!(da.map_data) || dbm_error(dbf)) break;
#endif
#endif

	}

	return values;
};


/*
**
*/

_ctb_map *
_ctb_map_collect(ctb_session *ctbsp, int fld) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;
	ssize_t mapsize = 0;
	ssize_t lastno = 0;
	unsigned int tmp;
	DBMAP *dbf;
	DBCHUNK av;
	DBCHUNK da;
	_ctb_map *map;

	if(!ctbs)
		return NULL;

	if(!(map=_ctb_findmap(ctbsp, fld)))
		return NULL;

	if(!(dbf=map->dbf)) {
		dbf = map->dbf = _ctb_map_open(map->mapfile, 0);

		map->OpenedReadOnly = 1;
		map->nentries = -1;
		map->lastindex = -1;

		if(!dbf)
			return NULL;
	};

	/* Cached number of entries */
	if(map->nentries != -1)
		return map;


#ifdef	HAVE_DBOPEN
	if(dbf->seq(dbf, &av, &da, R_FIRST) != 0)
		return NULL;
#else
#ifdef	HAVE_DBM_OPEN
		dbm_clearerr(dbf);
	av = dbm_firstkey(dbf);
		if(!(av.map_data) || dbm_error(dbf)) return NULL;
	da = dbm_fetch(dbf, av);
		if(!(da.map_data) || dbm_error(dbf)) return NULL;
#endif
#endif

	for(;;) {
		/* Got number! */
		if(av.map_size == (sizeof(tmp) + 1)
			&& *(char *)av.map_data == 0) {
			mapsize++;
			memcpy(&tmp, av.map_data + 1, sizeof(tmp));
			tmp = ntohl(tmp);
			if(tmp > lastno) lastno = tmp;
		}

#ifdef	HAVE_DBOPEN
		if(dbf->seq(dbf, &av, &da, R_NEXT) != 0)
			break;
#else
#ifdef	HAVE_DBM_OPEN
			dbm_clearerr(dbf);
		av = dbm_nextkey(dbf);
			if(!(da.map_data) || dbm_error(dbf)) break;
		da = dbm_fetch(dbf, av);
			if(!(da.map_data) || dbm_error(dbf)) break;
#endif
#endif
	}

	map->nentries = mapsize;
	map->lastindex = lastno;

	return map;
};


/*
** ctb_map_lastno()
**
** This function should be used to obtain the last map index.
**
*/

unsigned int
ctb_map_lastno(ctb_session *ctbsp, int fld) {
	_ctb_map *map;

	if(!(map = _ctb_map_collect(ctbsp, fld)))
		return 0;

	return map->lastindex;
};


ssize_t
ctb_map_size(ctb_session *ctbsp, int fld) {
	_ctb_map *map;

	if(!(map=_ctb_map_collect(ctbsp, fld)))
		return 0;

	return map->nentries;
};

