/*
** ctb_sbuf.c - a lite version of sbuf from libstrfunc.
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_STDARG_H
#include <stdarg.h>
#endif

#include "ctbint.h"

#define	INITIAL_SBUF_SIZE	16

_ctb_sbuf *
_ctb_sbuf_init() {
	register _ctb_sbuf *sb;
	sb = (_ctb_sbuf *)_ctb_calloc(1, sizeof(_ctb_sbuf));

	/* Initial fill */
	_ctb_sbuf_clear(sb);

	return sb;
};

int
_ctb_sbuf_clear(_ctb_sbuf *sb) {

	if(!sb)
		return -1;

	sb->size = INITIAL_SBUF_SIZE;
	sb->len = 0;
	sb->buf = (char *)_ctb_realloc(sb->buf, sb->size);

	*sb->buf = 0;

	return 0;
};

int
_ctb_sbuf_zero(_ctb_sbuf *sb) {
	if(!sb)
		return -1;

	sb->len = 0;

	if(sb->buf)
		*sb->buf = 0;
	else
		_ctb_sbuf_clear(sb);

	return 0;
};

void
_ctb_sbuf_free(_ctb_sbuf *sb) {
	if(!sb)
		return;

	if(sb->buf);
		free(sb->buf);

	free(sb);
};

#define	sizeround(foo,prevsize)	((foo)+((prevsize)-1))&(~(prevsize-1))

size_t
_ctb_sbuf_extend(_ctb_sbuf *sb, size_t newsize) {
	char *c;

	if(!sb)
		return -1;

	if(newsize >= sb->size)
		sb->buf = (char *)_ctb_realloc(sb->buf,
			sb->size = sizeround(newsize + 1, sb->size));

	/* Clear the rest of the buffer */
	c = sb->buf + sb->len;
	newsize = sb->size - sb->len;
	memset(c, 0, newsize);

	return sb->size;
};


ssize_t
_ctb_sbuf_add(_ctb_sbuf *sb, char *str) {
	register int size;

	if(!sb || !str)
		return -1;

	size = strlen(str);

	if (_ctb_sbuf_add2(sb, str, size) == -1)
		return -1;

	return size;
};

ssize_t
_ctb_sbuf_add2(_ctb_sbuf *sb, char *data, size_t size) {

	if(!sb || !data || (ssize_t)size < 0)
		return -1;

	if((sb->len + size) >= sb->size)
		sb->buf = (char *)_ctb_realloc(sb->buf,
			sb->size = sizeround((sb->len + size + 1), sb->size));

	memcpy(sb->buf + sb->len, data, size);

	sb->len += size;
	sb->buf[sb->len] = 0;

	return sb->len;
};


ssize_t
_ctb_sbuf_sprintf(_ctb_sbuf *sb, char *fmt, ...) {
	register size_t avail = 0;	/* Available buffer size */
	size_t size;
	va_list ap;

	if(!sb || !fmt)
		return -1;

	va_start(ap, fmt);

	if((avail = (sb->size - sb->len)) < 2) {
		_ctb_sbuf_extend(sb, sb->size + 2);
		avail = sb->size - sb->len;
	};

	while(avail <= (size=vsnprintf(sb->buf + sb->len, avail, fmt, ap))) {
		_ctb_sbuf_extend(sb, sb->size + size);
		avail = sb->size - sb->len;
	};

	sb->len += size;

	va_end(ap);

	return size;
};

char *
_ctb_sbuf_fgets(_ctb_sbuf *sb, FILE *fd) {
	register size_t avail=0;		/* Available buffer size */
	register size_t saved=-1;
	register char *p;

	if(!sb || !fd)
		return NULL;

  while(1) {

	if((avail = (sb->size - sb->len)) < 2) {
		_ctb_sbuf_extend(sb, sb->size + 2);
		avail = sb->size - sb->len;
	};

	p = fgets(sb->buf + sb->len, avail, fd);
	if(!p) {
		if(saved == -1)
			return NULL;
		return (sb->buf + saved);
	};

	if(saved == -1)
		saved = sb->len;

	sb->len += strlen(p);
	if( sb->buf[sb->len - 1] == '\n' )
		return sb->buf + saved;

  };	/* while(1) */

	return NULL;
};

