/*
** Author: Lev M. Walkin, 2000, 2001.
** E-mail: vlm@spelio.net.ru
** Web:    www.spelio.net.ru
**
** This code derived from libstrfunc project (http://www.spelio.net.ru/soft).
**
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

int
_ctb_initfdpos(const char *filename, int stream, const char *ext) {
	char *fpos;
	int posfd;
	off_t ofs=0;
	ino_t ino=0;
	time_t tloc;
	struct stat sb;


	if(!filename || !*filename || !ext || !*ext) {
		errno = EINVAL;
		return -1;
	};

	fpos=malloc(strlen(filename) + strlen(ext) + 2);
	if(!fpos)
		return -1;

	sprintf(fpos, "%s.%s", filename, ext);

	posfd = open(fpos, O_RDWR | O_CREAT, 0644);
	free(fpos);
	if(posfd == -1)
		return -1;

	if(lockf(posfd, F_TLOCK, sizeof(ino_t) + sizeof(off_t)))
		return -1;

	if(fstat(posfd, &sb)) {
		close(posfd);
		return -1;
	};
    if((sb.st_mode & S_IFMT) != S_IFREG) {
		close(posfd);
		return -1;
	};

	time(&tloc);

	lseek(posfd, 0, SEEK_SET);

	if(  (sb.st_size == 0)
	  ||  ( (sb.st_size != sizeof(ino_t) + sizeof(off_t))
	  && (sb.st_size % (sizeof(ino_t) + sizeof(off_t) + sizeof(time_t))) )
	) {

		fstat(stream, &sb);
		ino = sb.st_ino;
		ofs = 0;

		write(posfd, &ino, sizeof(ino_t));
		write(posfd, &ofs, sizeof(off_t));
		write(posfd, &tloc, sizeof(time_t));

	} else {
		/* If could not read basic info */
		if(read(posfd, &ino, sizeof(ino_t)) != sizeof(ino_t)
		|| read(posfd, &ofs, sizeof(off_t)) != sizeof(off_t))
			return -1;

		fstat(stream, &sb);

		if((ino != sb.st_ino) || (ofs > sb.st_size)) {
			ino = sb.st_ino;
			ofs = 0;
			lseek(posfd, 0, SEEK_SET);
			write(posfd, &ino, sizeof(ino_t));
			write(posfd, &ofs, sizeof(off_t));
			write(posfd, &tloc, sizeof(time_t));
		};
	};

	lseek(stream, ofs, SEEK_SET);

	return posfd;
};

int
_ctb_writeposfd(off_t offset, int posfd, int doSync) {

	if(offset < 0 || posfd == -1) {
		errno = EINVAL;
		return -1;
	};

	if(lseek(posfd, sizeof(ino_t), SEEK_SET) == -1)
		return -1;

	if(write(posfd, &offset, sizeof(off_t)) != sizeof(off_t))
		return -1;

	if(doSync) {
		time_t tloc;
		time(&tloc);

		write(posfd, &tloc, sizeof(time_t));

		fsync(posfd);
	};

	return 0;
};

int
_ctb_adjfdpos(int stream, int posfd, int doSync) {

	/* Determine current offset and write it to the position file */

	return _ctb_writeposfd(
		lseek(stream, 0, SEEK_CUR),
		posfd, doSync);

	return 0;
};

int
_ctb_initFILEpos(const char *filename, FILE *stream, const char *ext) {
	int posfd;
	off_t ofs;
	off_t ols;

	if(!stream) {
		errno = EINVAL;
		return -1;
	};

	ols = lseek(fileno(stream), 0, SEEK_CUR);

	posfd=_ctb_initfdpos(filename, fileno(stream), ext);
	if(posfd == -1)
		return -1;

	ofs = lseek(fileno(stream), 0, SEEK_CUR);
	if(ofs == -1)
		return -1;

	/* Rewind it where it should be. For transparency. */
	lseek(fileno(stream), ols, SEEK_SET);

	/* Advance a FILE position */
	if(fseek(stream, ofs, SEEK_SET) == -1)
		return -1;

	return posfd;
};

int
_ctb_adjFILEpos(FILE *stream, int posfd, int doSync) {
	off_t ofs;
	off_t newofs;
	long fofs;

	if(!stream) {
		errno=EINVAL;
		return  -1;
	};

	ofs = lseek(fileno(stream), 0, SEEK_CUR);
	fofs = ftell(stream);

	lseek(fileno(stream), fofs, SEEK_SET);

	if(_ctb_adjfdpos(fileno(stream), posfd, doSync) == -1)
		return -1;

	newofs = lseek(fileno(stream), 0, SEEK_CUR);

	/* Rewind it where it should be. For transparency. */
	lseek(fileno(stream), ofs, SEEK_SET);

	if(fseek(stream, newofs, SEEK_SET) == -1)
		return -1;

	return 0;
};



