#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "desire.h"
#include "session.h"
#include "ctbint.h"

char *_ctb_cf;
char *_ctb_cfpos;

extern int _ctb_lineno;
extern FILE *_ctb_in;

extern int _ctb_parse();

ctb_session *
ctb_open(const char *cfname, const char *service, const char *name)
{
	FILE *fp;
	struct stat sb;
	ctb_sess *ctbs;

	if(!service) {
		fprintf(stderr, "Service name required.\n");
		return NULL;
	};

	/* Create new session */
	ctbs = (ctb_sess *)_ctb_calloc(1, sizeof(ctb_sess));

	/* Fill initial service parameters */
	ctbs->service=_ctb_strdup(service);
	ctbs->name=name?_ctb_strdup(name):_ctb_strdup("");
	ctbs->EntrySize = sizeof(ctb_data);

	/* Open configuration file name */
	if(!cfname) cfname=CONFIG;

	if(
		!(fp=fopen(cfname, "r"))
		|| fstat(fileno(fp), &sb)
		|| ((sb.st_mode & S_IFMT) != S_IFREG)
	) {
		fprintf(stderr, "Can't open config file: %s\n", cfname);
		return NULL;
	};

	_ctb_lineno=1;

	_ctb_in = fp;

	if(_ctb_parse(ctbs)) {
		fprintf(stderr, "Config parse error: %s\n", cfname);
		ctb_close(ctbs);
		return NULL;
	};

	if(ctbs->ready == 0) {
		fprintf(stderr, "Service definition '%s' not found in %s\n",
			service, cfname);
		ctb_close(ctbs);
		return NULL;
	};

	fclose(fp);

	if(!ctbs->folder) {
		fprintf(stderr, "Folder not specified in config.\n");
		free(ctbs);
		return NULL;
	};

	if(stat(ctbs->folder, &sb)) {
		fprintf(stderr, "Folder '%s' does not exists.\n", ctbs->folder);
		free(ctbs);
		return NULL;
	};

	return (ctb_session *)ctbs;
}

void
ctb_close(ctb_session *ctbsp) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;
	_ctb_map *m1, *m2;

	if(!ctbs)
		return;

	if(ctbs->folder)
		free(ctbs->folder);
	if(ctbs->service)
		free(ctbs->service);
	if(ctbs->name)
		free(ctbs->name);
	if(ctbs->description)
		free(ctbs->description);

	for(m1=m2=ctbs->maps; m1; m1=m2) {
		m2=m1->next;
		if(m1->mapfile)
			free(m1->mapfile);

#ifdef	HAVE_DBOPEN
		if(m1->dbf)
			m1->dbf->close(m1->dbf);
		m1->dbf = NULL;
#else
#ifdef	HAVE_DBM_OPEN
		if(m1->dbf)
			dbm_close(m1->dbf);
		m1->dbf = NULL;
#endif
#endif

		free(m1);
	};

	if(ctbs->cache_add_name)
		free(ctbs->cache_add_name);

	if(ctbs->cache_add_fp) {
		fflush(ctbs->cache_add_fp);
		fclose(ctbs->cache_add_fp);
	};

	free(ctbs);
};

void
ctb_sync(ctb_session *ctbsp) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;
	_ctb_map *map;

	for(map=ctbs->maps; map; map=map->next) {
		if(!map->dbf)
			continue;
#ifdef	HAVE_DBOPEN
		map->dbf->sync(map->dbf, 0);
#else
#ifdef	HAVE_DBM_OPEN
		// dbm_sync?(map->dbf);
#endif
#endif
	}

	if( ctbs->cache_add_fp )
		fflush(ctbs->cache_add_fp);

}

char *
ctb_description(ctb_session *ctbsp) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;

	if(!ctbs)
		return NULL;

	if(ctbs->description)
		return ctbs->description;
	if(ctbs->name)
		return ctbs->name;
	if(ctbs->service)
		return ctbs->service;

	return "unknown service";
};

unsigned int
ctb_entry_size(ctb_session *ctbsp) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;

	if(!ctbs)
		return 32;

	return ctbs->EntrySize;
};

int
ctb_chname(ctb_session *ctbsp, const char *name) {
	register ctb_sess *ctbs = (ctb_sess *)ctbsp;

	if(!ctbs)
		return 0;

	if(ctbs->name)
		free(ctbs->name);
	ctbs->name=_ctb_strdup(name?name:"");

	if(ctbs->description) {
		free(ctbs->description);
		ctbs->description=NULL;
	};

	if(ctbs->cache_add_name) {
		free(ctbs->cache_add_name);
		ctbs->cache_add_name=NULL;
	};

	if(ctbs->cache_add_fp) {
		fclose(ctbs->cache_add_fp);
		ctbs->cache_add_fp=NULL;
	};

	return 0;
};

